/* Test software for gjrand random numbers version 3.4.1.0 or later. */
/* Copyright (C) 2004-2011 G. Jones. */
/* Licensed under the GNU General Public License version 2 or 3. */

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "../../testcommon/pcombo.h"
#include "../../testcommon/binom.h"
#include "../../testcommon/chi2p.h"

/* Nearest distance analysis. */
/* Like Kendall and Babington Smith Gap Test but bigger and less elegant. */
/* Culminates in a huge Chi-Squared test. */
/* This is Geronimo's own invention (dating from February 2005), though */
/* someone else may have done it before. Please mention in gjrand forums on */
/* SourceForge if this already has a name. */

/* Random double precision reals in native binary format are read from */
/* standard input. There is an optional size arg in bytes, otherwise it */
/* just reads until end of file. */

static void
crash(char *s)
{
    fprintf(stderr, "crash [%s]\n", s); exit(1);
}

static int errorlevel=0;
static void seterr(int e) {if (errorlevel<e) errorlevel=e;}

static double dc[16][16][50];
static int64_t count[16][16][50];
static int32_t c32[16][16][50];
static int32_t last[16];

static void
init(void)
{
    int i;
    memset(count, 0, sizeof(count));
    memset(c32, 0, sizeof(c32));
    for (i=0; i<16; i++) last[i]= -11;
}

static void
donibble(const double *b, int pos)
{
    double w, x, y, z;
    int j, n;
    int32_t *c;

    w = b[0]; x = b[1]; y = b[2]; z = b[3];
    n = (w>x) + 2*(y>z);
    w += x; y += z;
    n = (w>y) + 2*n;
    n = (w+y>2.0) + 2*n;

    c = &(c32[n][0][0]);

    for (j= -16; j<0; j++)
    {
	int d=pos-(last+16)[j];
	if (d>49) d=49;
	c[d]++;
	c+=50;
    }

    last[n]=pos;
}

static void
dobuf(const double *buf, int pos, int i)
{
    int j;

    buf+=4*i;

    for (j= -4*i; j<0; j+=4)
    {
	donibble(buf+j, pos);
	pos++;
    }
}

static double expect[50];

static void
mkexpect(int64_t size)
{
    int i;
    double p=size/16.0, x;

    for (i=1; i<=48; i++)
    {
	x=p*(1.0/16.0);
	expect[i]=x;
	p-=x;
    }
    expect[49]=p;
}

static void
butterfly(double *x, double *y)
{
    double a, b;
    int siz=y-x;

    do
    {
	a = *x; b = *y;
	*x = a+b; *y = a-b;
	x++; y++;
	siz--;
    } while (siz>0);

    siz = y-x;
    if (siz>50)
    {
	x -= siz; y -= siz;
	siz >>= 1;
	butterfly(x, x+siz);
	butterfly(y, y+siz);
    }
}

static time_t tstart;

static void
doan(int64_t n, int final)
{
    double p[3];
    double t=0.0, extreme;
    double x;
    time_t tm;
    int exi, exj, exk, exd='?';
    int i, j, k;

    mkexpect(n);

    for (i=0; i<16; i++) for (j=0; j<16; j++) for (k=1; k<=49; k++)
    {
	double x=(double)(count[i][j][k]);
	double e=expect[k];
	x-=e;
	t+=x*x/e;
    }

    printf("%15.5f\n", (double)t);

    x = (t - 12761) * ((82.001-60.332)/(13379-12761)) + 60.332;
    if (x<0.0) x = 0.0;
    p[0] = chi2p1(x, 52);

    t=0.0; extreme= 999.0; exi=exj=exk= -1;
    for (i=0; i<16; i++) for (j=0; j<16; j++)
    {
	uint64_t tot=0;
	for (k=49; k>0; k--) tot+=count[i][j][k];
	for (k=1; k<=48; k++)
	{
	    uint64_t x=count[i][j][k];
	    double y=sumbino(tot, x, 1.0/16.0);
	    double z = tot*(1.0/16.0);

	    if (y<extreme)
	    {
		extreme=y; exi=i; exj=j; exk=k;
		if (x>z) exd='+'; else exd='-';
	    }

	    dc[i][j][k] = (x-z) / sqrt(z);

	    tot-=x;
	}
    }

    printf("extreme = %.3g   (%d %d %d %c)\n", extreme, exi, exj, exk, exd);
    p[1]=pco_scale(extreme, 16*16*48);

    butterfly(&(dc[0][0][0]), &(dc[8][0][0]));
    extreme = -1.0;
    for (i=0; i<16; i++) for (j=0; j<16; j++) for (k=1; k<=48; k++)
    {
	t = fabs(dc[i][j][k]);
	if (t>extreme) {extreme = t; exi = i; exj = j; exk = k;}
    }

    extreme /= 16.0;
    printf("transform = %.3g   (%x %x %d)\n", extreme, exi, exj, exk);

    extreme = erfc(M_SQRT1_2*extreme);
    p[2] = pco_scale(extreme, 16*16*48);

    printf("pvals (%.3g %.3g %.3g)\n", p[0], p[1], p[2]);

    tm = time(0);
    printf("\nprocessed %.2g numbers in %.3g seconds. %s\n",
        n*4.0, (double)(tm-tstart), ctime(&tm));

    printf("one sided P value (very small numbers are bad)\n");
    printf("%c = %.3g\n", "pP"[final], pcombo(p, 3));
    if (!final) printf("------\n\n");
}

static void
adjust(int pos)
{
    int i, j, k;

    for (i=0; i<16; i++)
    {
	j=last[i]-pos;
	if (j< -9000000) j= -9000000;
	last[i]=j;
    }

    for (i=0; i<16; i++) for (j=0; j<16; j++) for (k=1; k<=49; k++)
    {
	count[i][j][k]+=c32[i][j][k];
	c32[i][j][k]=0;
    }
}

static void
readstuff(int64_t n, int progress)
{
    static int64_t progsize[]=
    {
	500000,
	700000,
	1000000,
	1500000,
	2000000,
	3000000,
	0
    };
    double buf[1024];
    int64_t posll=0, nextp=progsize[0];
    int pos=0, j, progi=0;

    n >>= 2;

    while (n<0 || posll<n)
    {
	j=fread(buf, 4*sizeof(double), 256, stdin);
	if (j<=0) break;
	if (n>0 && j>n-posll) j=n-posll;
	dobuf(buf, pos, j);
	posll+=j; pos+=j;
	if (pos>=4000000) {adjust(pos); pos=0;}
	if (progress && 4*posll>nextp)
	{
	    if (pos) {adjust(pos); pos=0;}
	    doan(posll, 0);
	    progsize[progi] *= 10;
	    progi++;
	    nextp = progsize[progi];
	    if (!nextp) {progi = 0; nextp = progsize[progi];}
	}
    }

    adjust(pos);

    if (n>0 && posll<n)
    {
fprintf(stderr, "Warning, expected %.0f 8 double samples, saw only %.0f\n",
		(double)n, (double)posll);
	seterr(1);
    }

    if (posll<100000)
    {
	fprintf(stderr, "Warning, < 100000 samples, don't trust result.\n");
	seterr(1);
    }

    doan(posll, 1);
}

int
main(int argc, char **argv)
{
    double dn;
    int64_t n= -1;
    int progress=0, j;

    tstart = time(0);

    for (j=1; j<argc; j++)
    {
	if (strcmp(argv[j], "--progress")==0) progress = 1;
	else if (sscanf(argv[j], "%lf", &dn)==1) n = (int64_t)dn;
	else crash("optional arg must be --progress or numeric");
    }

    /* although we do a chi-square test, when tested on several random */
    /* generators i trust a lot, the results do not follow the chi-square */
    /* distribution. The mean is about right near 12520, but the standard */
    /* deviation is near 285 compared to about 150 for the chi-square */
    /* distribution. It may be that this is because the counts we are */
    /* studying correlate with each other in different ways than is usually */
    /* the case with the chi-square test. These expected values are obtained */
    /* emperically and are only approximate. Theoretical analysis is invited. */

    init();
    printf("expected range [ 11770 11940 12170 12890 13270 13570 ]\n");

    readstuff(n, progress);

    return errorlevel;
}
