/* Test software for gjrand random numbers version 3.4.1.0 or later. */
/* Copyright (C) 2004-2011 G. Jones. */
/* Licensed under the GNU General Public License version 2 or 3. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "../../src/my_int_types.h"
#include "../../testcommon/pcombo.h"
#include "../../testcommon/binom.h"
#include "../../testcommon/chi2p.h"

/* "Kendall and Babington-Smith Gap Test" on double precision values. */

/* For each number read, calculate the distance since the same number */
/* value was last seen and collect a histogram of these distances. */
/* Finally do a chi-square test on the histogram. */

static int errorlevel=0;
static void seterr(int e) {if (errorlevel<e) errorlevel=e;}

static void
crash(char *s)
{
    fprintf(stderr, "crash [%s]\n", s); exit(1);
}

#define MULT 6827
#define MAXLAG 20000
#define ZDF 13535

static int64_t dc[3][MAXLAG];
static int64_t pos64[MULT];

static void
mkexpect(double expect[MAXLAG], int64_t pos)
{
    double c=(double)pos, e=c/MULT;
    int i;

    for (i=1; i<MAXLAG; i++)
    {
	expect[i]=e;
	c-=e;
	e *= (MULT-1.0)/MULT;
    }

    expect[0]=c;
}


static void
doan1(const char *note, int64_t n, int j, int df, double pvs[2])
{
    double expect[MAXLAG];
    double e, t=0.0, x, extreme;
    int64_t tot;
    int i, exi, exd, tries=0;

    if (n<44000) {pvs[0] = pvs[1] = 0.5; return;}

    mkexpect(expect, n);

    for (i=0; i<MAXLAG; i++)
    {
	e=expect[i];
	x=(double)dc[j][i]-e;
	t+=x*x/e;
    }

    pvs[0] = chi2p1(t, df);

    printf("%-9s %10.2f (p = %.3g) : ", note, t, pvs[0]);

    extreme=999.0; exi= -1; exd='?';
    if (pvs[0]<=0.0) {pvs[1] = 0.5; printf("-\n"); return;}

    if (df<MAXLAG-10)
    {
	/* I don't understand the distribution very well. This analysis */
	/* is very approximate and tries to be conservative. ie if a low */
	/* P-value results, the true P-value is probably even lower. */
	tot=n;
	for (i=3; i<MAXLAG-2; i++)
	{
	    double xp;
	    int64_t k=dc[j][i];
	    xp = sqrt((double)k)*dc[j][i-2]*dc[j][i-1]*dc[j][i+1]*dc[j][i+2];
	    if (tot>3 && xp>280.7412308013823)
	    {
		xp = pow(xp, 2.0/9.0);
		tries++;
		x=sumbino(tot, k, xp/tot);
		if (x<extreme)
		{
		    extreme=x; exi=i;
		    if (k*MULT>tot) exd='+'; else exd='-';
		}
	    }
	    tot-=k;
	}
    }
    else
    {
	tot=n;
	for (i=1; i<MAXLAG && extreme>0; i++)
	{
	    int64_t k=dc[j][i];

	    if (tot>3)
	    {
		tries++;
		x=sumbino(tot, k, 1.0/MULT);
		if (x<extreme)
		{
		    extreme=x; exi=i;
		    if (k*MULT>tot) exd='+'; else exd='-';
		}
	    }
	    tot-=k;
	}
    }

    pvs[1] = tries ? pco_scale(extreme, tries) : 0.5;

    printf("extreme = %.3g (%d %c) (p = %.3g)\n", extreme, exi, exd, pvs[1]);
}

static time_t tstart;

static void
doan(int64_t n, int final)
{
    double pvs[6];
    int64_t p;
    time_t tm;
    int j;

    doan1("simple", n, 0, MAXLAG-2, pvs);
    doan1("half", n, 1, MAXLAG-2, pvs+2);
    p = 0;
    for (j=0; j<MAXLAG; j++) p += dc[2][j];
    doan1("fromzero", p, 2, ZDF, pvs+4);

    tm = time(0);
    printf("\nprocessed %.2g numbers in %.3g seconds. %s\n",
        (double)n, (double)(tm-tstart), ctime(&tm));

    if (!final) printf("progress ");

    printf("one sided P value (very small numbers are bad)\n");
    printf("%c = %.3g\n", "pP"[final], pcombo(pvs, 6));
    if (!final) printf("------\n\n");
}

static void
dobuf(const double *buf, int64_t p, int n)
{
    int64_t j;
    const double * const bend=buf+n;
    int i, ch, k;

    for (i= -n; i<0; i++)
    {
	ch = (int)(MULT*bend[i]);

	j = p-pos64[ch];
	if (j>=MAXLAG) j = 0;
	dc[0][j]++;

	j = p-pos64[ch>>1];
	if (j>=MAXLAG) j = 0;
	dc[1][j]++;

	if (ch==0)
	{
	    for (k=MULT-1; k>=0; k--)
	    {
		j = p-pos64[k];
		if (j>=MAXLAG) j = 0;
		dc[2][j]++;
	    }
	}

	pos64[ch] = p; p++;
    }
}

static void
readstuff(int64_t n, int progress)
{
    static int64_t progsize[]=
    {
	500000,
	700000,
	1000000,
	1500000,
	2000000,
	3000000,
	0
    };
    double buf[1024];
    int64_t p=0, nextp=progsize[0];
    int j, progi=0;

    for (j=0; j<MULT; j++) pos64[j]= -1-j;
    memset(dc, 0, sizeof(dc));

    while (n<=0 || p<n)
    {
	j=fread(buf, sizeof(double), 1024, stdin);
	if (j<=0) break;
	if (n>0 && j>n-p) j = n-p;
	dobuf(buf, p, j);
	p += j;
	if (progress && p>nextp)
	{
	    doan(p, 0);
	    progsize[progi] *= 10;
	    progi++;
	    nextp = progsize[progi];
	    if (!nextp) {progi = 0; nextp = progsize[progi];}
	}
    }

    if (n>0 && p<n)
    {
	fprintf(stderr, "Warning, expected %.0f numbers, saw only %.0f\n",
		(double)n, (double)p);
	seterr(1);
    }

    if (p<450000)
    {
	fprintf(stderr, "Warning, < 450000 samples, don't trust result.\n");
	seterr(1);
    }

    doan(p, 1);
}

int
main(int argc, char **argv)
{
    double dn;
    int64_t n= -1;
    int progress=0, j;

    tstart = time(0);

    for (j=1; j<argc; j++)
    {
	if (strcmp(argv[j], "--progress")==0) progress = 1;
	else if (sscanf(argv[j], "%lf", &dn)==1) n = (int64_t)dn;
	else crash("optional arg must be --progress or numeric");
    }

    readstuff(n, progress);

    return errorlevel;
}
