/* Test software for gjrand random numbers version 3.4.1.0 or later. */
/* Copyright (C) 2004-2011 G. Jones. */
/* Licensed under the GNU General Public License version 2 or 3. */

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "../../src/my_int_types.h"
#include "../../testcommon/pcombo.h"
#include "../../testcommon/binom.h"
#include "../../testcommon/chi2p.h"

/* "Kendall and Babington-Smith Gap Test" on 4 bit parts of 32 bit words. */
/* This is most likely to be useful on generators which produce 32 bit words */
/* with some bits much worse quality than others. I am thinking mostly */
/* of linear congruential generators, but perhaps it will catch others too. */

static void
crash(char *s)
{
    fprintf(stderr, "crash [%s]\n", s); exit(1);
}

static int errorlevel=0;
static void seterr(int e) {if (errorlevel<e) errorlevel=e;}

static int64_t dc[12][51];
static int pos32[12][16];

static void
mkexpect(double expect[51], int64_t pos)
{
    double c=(double)pos, e=c/16.0;
    int i;

    for (i=1; i<50; i++)
    {
	expect[i]=e;
	c-=e;
	e *= 15.0/16.0;
    }

    expect[50]=c;
}

static time_t tstart;

static void
doan(int64_t n, int final)
{
    static const int shifts[12]={28, 26, 24,  20, 18, 16,  12, 10, 8,  4, 2, 0};
    double expect[51];
    double p[12], extreme;
    time_t tm;
    int i, j, exi, exj, exd;

    mkexpect(expect, n);

    for (j=0; j<12; j++)
    {
	int64_t *d= &(dc[j][0]);
	double e, x, t=0.0;
	for (i=1; i<51; i++)
	{
	    e=expect[i];
	    x=(double)d[i]-e;
	    t+=x*x/e;
	}
	printf("%2d : %15.5f\n", shifts[j], t);

	p[j] = chi2p1(t, 49);
    }

    p[0]=pcombo(p, 12);

    extreme=999.0; exi= -1; exj= -1; exd='?';
    for (j=0; j<12; j++)
    {
	int64_t tot=n;
	for (i=1; i<50; i++)
	{
	    int64_t k=dc[j][i];
	    double x=sumbino(tot, k, 1.0/16.0);
	    if (x<extreme)
	    {
		extreme=x; exi=i; exj=j;
		if (k*16>tot) exd='+'; else exd='-';
	    }
	    tot-=k;
	}
    }
    printf("extreme = %.3g (%d %d %c)\n", extreme, shifts[exj], exi, exd);
    p[1]=pco_scale(extreme, 12*49);

    printf("pvals (%.3g %.3g)\n\n", p[0], p[1]);

    tm = time(0);
    printf("\nprocessed %.2g bytes in %.3g seconds. %s\n",
	n*4.0, (double)(tm-tstart), ctime(&tm));

    if (!final) printf("progress ");
    printf("one sided P value (very small numbers are bad)\n");
    printf("%c = %.3g\n", "pP"[final], pcombo(p, 2));
    if (!final) printf("------\n\n");
}

static void
dobuf(const uint32_t * const buf, int n)
{
    int i, j, k;
    const uint32_t * const bend=buf+n;

    for (k=0; k<12; k++)
    {
	for (i=0; i<16; i++)
	{
	    j=pos32[k][i]-n;
	    if (j< -20000) j= -20000;
	    pos32[k][i]=j;
	}
    }

    i= -n;
    do
    {
	uint32_t ch=bend[i];

#define DOPART(k,sh) \
	{ \
	    int part=(ch>>sh)&15; \
	    j=i-pos32[k][part]; \
	    if (j>50) j=50; \
	    pos32[k][part]=i; \
	    dc[k][j]++; \
	}

	DOPART(0, 28);
	DOPART(1, 26);
	DOPART(2, 24);
	DOPART(3, 20);
	DOPART(4, 18);
	DOPART(5, 16);
	DOPART(6, 12);
	DOPART(7, 10);
	DOPART(8, 8);
	DOPART(9, 4);
	DOPART(10, 2);
	DOPART(11, 0);

	i++;
    }
    while (i<0);
}

static void
readstuff(int64_t n, int progress)
{
    static int64_t progsize[]=
    {
	100000000,
	150000000,
	200000000,
	300000000,
	500000000,
	700000000,
	0
    };
#define MYBUFSIZ 1024
    uint32_t buf[MYBUFSIZ];
    int64_t p=0, nextp=progsize[0];
    int j, progi=0;

    memset(dc, 0, sizeof(dc));
    for (j=0; j<12; j++)
    {
	int i;
	for (i=0; i<16; i++) pos32[j][i]= -18;
    }

    while (n<0 || p<n)
    {
	j=fread(buf, sizeof(uint32_t), MYBUFSIZ, stdin);
	if (j<=0) break;
	if (n>0 && j>n-p) j=n-p;
	dobuf(buf, j);
	p+=j;
	if (progress && p*4>nextp)
	{
	    doan(p, 0);
	    progsize[progi] *= 10;
	    progi++;
	    nextp = progsize[progi];
	    if (!nextp) {progi = 0; nextp = progsize[progi];}
	}
    }

    if (n>0 && p<n)
    {
	fprintf(stderr,
		"Warning, expected %.0f 32 bit samples, saw only %.0f\n",
		(double)n, (double)p);
	seterr(1);
    }

    if (p<1000)
    {
	fprintf(stderr, "Warning, < 1000 samples, don't trust result.\n");
	seterr(1);
    }

    doan(p, 1);
}

int
main(int argc, char **argv)
{
    double dn;
    int64_t n= -1;
    int progress=0, j;

    tstart = time(0);

    for (j=1; j<argc; j++)
    {
	if (strcmp(argv[j], "--progress")==0) progress = 1;
	else if (sscanf(argv[j], "%lf", &dn)==1) n = (int64_t)(dn*0.25);
	else crash("optional arg must be --progress or numeric");
    }

    printf("expected range [ 23.8 28.7 36.6 62.0 75.0 85.9 ]\n");

    readstuff(n, progress);

    return errorlevel;
}
