/* Test software for gjrand random numbers version 3.4.1.0 or later. */
/* Copyright (C) 2004-2011 G. Jones. */
/* Licensed under the GNU General Public License version 2 or 3. */

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "../../src/my_int_types.h"
#include "../../testcommon/pcombo.h"
#include "../../testcommon/binom.h"
#include "../../testcommon/chi2p.h"

/* "Kendall and Babington-Smith Gap Test" on byte values. */

/* For each byte read, calculate the distance since the same byte */
/* value was last seen and collect a histogram of these distances. */
/* Finally do a chi-square test on the histogram. */

static void
crash(char *s)
{
    fprintf(stderr, "crash [%s]\n", s); exit(1);
}

static int errorlevel=0;
static void seterr(int e) {if (errorlevel<e) errorlevel=e;}

static int64_t dc[3][752];
static int pos32[256];

static void
mkexpect(double expect[752], int64_t pos)
{
    double c=(double)pos, e=c/256.0;
    int i;

    for (i=1; i<751; i++)
    {
	expect[i]=e;
	c-=e;
	e *= 255.0/256.0;
    }

    expect[751]=c;
}


static void
doan1(const char *note, int64_t n, int j, int df, double pvs[2])
{
    double expect[752];
    double e, t=0.0, x, extreme;
    int64_t tot;
    int i, exi, exd;

    mkexpect(expect, n);

    for (i=1; i<752; i++)
    {
	e=expect[i];
	x=(double)dc[j][i]-e;
	t+=x*x/e;
    }

    pvs[0] = chi2p1(t, df);

    printf("%-9s %10.2f (p = %.3g)\n", note, t, pvs[0]);

    if (df<700) return;

    extreme=999.0; exi= -1; exd='?';
    tot=n;
    for (i=1; i<751; i++)
    {
	int64_t k=dc[j][i];
	x=sumbino(tot, k, 1.0/256.0);
	if (x<extreme)
	{
	    extreme=x; exi=i;
	    if (k*256>tot) exd='+'; else exd='-';
	}
	tot-=k;
    }
    pvs[1]=pco_scale(extreme, df+1);
    printf("extreme = %.3g (%d %c) (p = %.3g)\n\n", extreme, exi, exd, pvs[1]);
}

static time_t tstart;

static void
doan(int64_t n, int final)
{
    double pvs[6];
    int64_t p;
    time_t tm;
    int j;

    doan1("simple", n, 0, 749, pvs);
    doan1("half", n, 1, 749, pvs+2);
    p = 0;
    for (j=0; j<752; j++) p += dc[2][j];
    doan1("fromzero", p, 2, 508, pvs+4);

    tm = time(0);
    printf("\nprocessed %.2g bytes in %.3g seconds. %s\n",
	(double)n, (double)(tm-tstart), ctime(&tm));

    if (!final) printf("progress ");
    printf("one sided P value (very small numbers are bad)\n");
    printf("%c = %.3g\n", "pP"[final], pcombo(pvs, 5));
    if (!final) printf("------\n\n");
}

static void
dobuf(const unsigned char *buf, int n)
{
    int i, j, ch, k;
    const unsigned char * const bend=buf+n;

    for (k=255; k>=0; k--)
    {
	j = pos32[k]-n;
	if (j < -20000) j = -20000;
	pos32[k] = j;
    }

    i = -n;
    do
    {
	ch = bend[i];

	j = i-pos32[ch];
	if (j>751) j = 751;
	dc[0][j]++;

	j = i-pos32[ch>>1];
	if (j>751) j = 751;
	dc[1][j]++;

	if (ch==0)
	{
	    for (k=255; k>=0; k--)
	    {
		j = i-pos32[k];
		if (j>751) j = 751;
		dc[2][j]++;
	    }
	}

	pos32[ch] = i;
	i++;
    }
    while (i<0);
}

static void
readstuff(int64_t n, int progress)
{
    static int64_t progsize[]=
    {
	100000000,
	150000000,
	200000000,
	300000000,
	500000000,
	700000000,
	0
    };
#define MYBUFSIZ 8192
    unsigned char buf[MYBUFSIZ];
    int64_t p=0, nextp=progsize[0];
    int j, progi=0;

    for (j=0; j<256; j++) pos32[j]= -9-j;
    memset(dc, 0, sizeof(dc));

    while (n<=0 || p<n)
    {
	j=fread(buf, 1, MYBUFSIZ, stdin);
	if (j<=0) break;
	if (n>0 && j>n-p) j = n-p;
	dobuf(buf, j);
	p += j;
	if (progress && p>nextp)
	{
	    doan(p, 0);
	    progsize[progi] *= 10;
	    progi++;
	    nextp = progsize[progi];
	    if (!nextp) {progi = 0; nextp = progsize[progi];}
	}
    }

    if (n>0 && p<n)
    {
	fprintf(stderr, "Warning, expected %.0f samples, saw only %.0f\n",
		(double)n, (double)p);
	seterr(1);
    }

    if (p<500000)
    {
	fprintf(stderr, "Warning, < 500000 samples, don't trust result.\n");
	seterr(1);
    }

    doan(p, 1);
}

int
main(int argc, char **argv)
{
    double dn;
    int64_t n= -1;
    int progress=0, j;

    tstart = time(0);

    for (j=1; j<argc; j++)
    {
	if (strcmp(argv[j], "--progress")==0) progress = 1;
	else if (sscanf(argv[j], "%lf", &dn)==1) n = (int64_t)dn;
	else crash("optional arg must be --progress or numeric");
    }

    printf("expected range [ 633 662 699 799 841 875 ]\n");

    readstuff(n, progress);

    return errorlevel;
}
