/* Test software for gjrand random numbers version 3.4.1.0 or later. */
/* Copyright (C) 2004-2011 G. Jones. */
/* Licensed under the GNU General Public License version 2 or 3. */

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "../../src/my_int_types.h"
#include "../../testcommon/pcombo.h"

/* Bytes of input are interpreted as signed numbers (two schemes are */
/* available with -f or -r). Block size currently 64kB. */
/* We add the same byte position from each block to get a 64k sized */
/* block of totals. Then do an autocorrelation for this totals block */
/* for all lags up to 32k. It seems to me that this should be less powerful */
/* than doing autocorrelation seperately on each block and then analysing */
/* all those results (RMS?). On the other hand, this method is hugely faster. */
/* Even using FFT for autocorrelation wouldn't make it very practical to */
/* seperately autocorrelate each block for GigaBytes. */

static void
crash(char *s)
{
    fprintf(stderr, "crash [%s]\n", s); exit(1);
}

static int errorlevel=0;
static void seterr(int e) {if (errorlevel<e) errorlevel=e;}

#define SHIFT 16
#define SIZE (1<<SHIFT)

#define MAXMEM (SIZE*11)

static double memory[MAXMEM];
static double *memptr=memory;

static short tmpacc[2*SIZE];
static int tmpsat=0;

static double *
memmark(int siz)
{
    double *r=memptr;

    memptr=r+siz;
if (memptr>memory+MAXMEM) crash("out of memory");

    return r;
}

static void
memrelease(double *mark)
{
    memptr=mark;
}


struct poly {int order; double *coef;};

static void
polysub2(double *ca, double *cr, int ord)
{
    int i=ord-2;

    do {cr[1+i]-=ca[1+i]; cr[i]-=ca[i]; i-=2;} while (i>=0);
}

static void
polyadd2(double *ca, double *cr, int ord)
{
    int i=ord-2;

    do {cr[1+i]+=ca[1+i]; cr[i]+=ca[i]; i-=2;} while (i>=0);
}

static void
polyadd3(double *ca, double *cb, double *cr, int ord)
{
    int i=ord-2;

    do {cr[1+i]=ca[1+i]+cb[1+i]; cr[i]=ca[i]+cb[i]; i-=2;} while (i>=0);
}

static void
polymul4(double *ca, double *cb, double *cr)
{
    cr[0]=ca[0]*cb[0];
    cr[1]=ca[0]*cb[1]+ca[1]*cb[0];
    cr[2]=ca[0]*cb[2]+ca[1]*cb[1]+ca[2]*cb[0];
    cr[3]=ca[0]*cb[3]+ca[1]*cb[2]+ca[2]*cb[1]+ca[3]*cb[0];
    cr[4]=ca[1]*cb[3]+ca[2]*cb[2]+ca[3]*cb[1];
    cr[5]=ca[2]*cb[3]+ca[3]*cb[2];
    cr[6]=ca[3]*cb[3];
    cr[7]=0;
}

static void
polymul2(double *ca, double *cb, double *cr, int ord)
{
    if (ord==4) {polymul4(ca, cb, cr); return;}

    {
	int i;
	struct poly t1, t3;

	for (i=0; i<2*ord; i++) cr[i]=0;

	t1.order=ord; t3.order=ord;
	t1.coef=memmark(ord);

	polymul2(ca, cb, t1.coef, ord/2);
	polyadd2(t1.coef, cr, ord);
	polysub2(t1.coef, cr+ord/2, ord);

	polymul2(ca+ord/2, cb+ord/2, t1.coef, ord/2);
	polyadd2(t1.coef, cr+ord, ord);
	polysub2(t1.coef, cr+ord/2, ord);

	polyadd3(ca, ca+ord/2, t1.coef, ord/2);
	polyadd3(cb, cb+ord/2, t1.coef+ord/2, ord/2);

	t3.coef=memmark(ord);
	polymul2(t1.coef, t1.coef+ord/2, t3.coef, ord/2);

	polyadd2(t3.coef, cr+ord/2, ord);

	memrelease(t1.coef);
    }
}

static struct poly
polymul(struct poly a, struct poly b)
{
    struct poly r;

    if (a.order!=b.order) crash("polymul order mismatch");
    r.order=2*a.order;
    r.coef=memmark(r.order);
    polymul2(a.coef, b.coef, r.coef, a.order);

    return r;
}

static struct poly
revpoly(struct poly p)
{
    struct poly r;
    int i;

    r.order=p.order;
    r.coef=memmark(p.order);

    for (i=0; i<p.order; i++) r.coef[i]=p.coef[p.order-i-1];

    return r;
}

static void
zeropoly(struct poly r)
{
    int i;
    for (i=0; i<SIZE; i++) r.coef[i]=0.0;
}

static double accstats[SIZE/2];

static void
dostats(struct poly p)
{
    int i;

if (p.order!=2*SIZE) crash("dostats bad p.order");

    for (i=0; i<SIZE/2; i++) accstats[i]=p.coef[SIZE+i]+p.coef[i];
}

static double
printstats(int64_t bcount)
{
    int i, j, count;
    double t, x, low[3], high[3], le;
    int ll[3], hl[3];
    double s=0.0, ss=0.0, scale=(sqrt(SIZE)/3.0/21845.0/21845.0)/bcount;

    for (j=0; j<=2; j++) {low[j] = 1e38; high[j] = -1e38; ll[j] = hl[j] = -1;}

    count = SIZE/2;

    for (i=0; i<count; i++)
    {
	j=i+1;

	x=accstats[i]*scale;

	if (x>high[2])
	{
	    high[2]=x; hl[2]=j;
	    if (x>high[1])
	    {
		high[2]=high[1]; hl[2]=hl[1];
		high[1]=x; hl[1]=j;
		if (x>high[0])
		{
		    high[1]=high[0]; hl[1]=hl[0];
		    high[0]=x; hl[0]=j;
		}
	    }
	}
	if (x<low[2])
	{
	    low[2]=x; ll[2]=j;
	    if (x<low[1])
	    {
		low[2]=low[1]; ll[2]=ll[1];
		low[1]=x; ll[1]=j;
		if (x<low[0])
		{
		    low[1]=low[0]; ll[1]=ll[0];
		    low[0]=x; ll[0]=j;
		}
	    }
	}
	s+=x; ss+=x*x;
    }

    printf(" : mean=%f", s/count);
    printf(" ; sd  =%f\n", sqrt(ss/count-s*s/((double)count*(double)count)));

    /* Somewhere here i suspect my probability calculations are wrong. */
    /* On the high side, they seem to be somewhere near right. On the */
    /* low side, very low numbers seem to be much rarer than these */
    /* figures suggest. Perhaps one of my dubious assumptions about */
    /* normality and independence is biting. Never mind. If a generator */
    /* blows out big-time (P<1.0e-6) you can rest assured it has done */
    /* something unlikely and suspicious, even if the exact P-value is */
    /* not quite right. */

    printf("high=%f (%d)  %f (%d)  %f (%d)\n",
	high[0], hl[0], high[1], hl[1], high[2], hl[2]);
    printf("low =%f (%d)  %f (%d)  %f (%d)\n",
	low[0], ll[0], low[1], ll[1], low[2], ll[2]);

    t = -low[0];
    if (high[0]>t) t=high[0];

    le = erfc(M_SQRT1_2*t)*0.5;
    le = pco_scale(le, (double)SIZE);
    printf("==> p = %.3g\n", (double)le);

    return le;
}

static time_t tstart;

static void
doan(int64_t n, int final, struct poly accf, struct poly accr)
{
    struct poly q, m;
    double pvs[2], *mark;
    time_t tm;

    mark = memmark(1);
    q=revpoly(accr);
    m=polymul(accr, q);
    dostats(m);
    printf("reverse");
    pvs[0] = printstats(n);
    memrelease(mark);

    mark = memmark(1);
    q=revpoly(accf);
    m=polymul(accf, q);
    dostats(m);
    printf("forward");
    pvs[1] = printstats(n);
    memrelease(mark);

    tm = time(0);
    printf("\nprocessed %.2g bytes in %.3g seconds. %s\n",
	(double)n*SIZE, (double)(tm-tstart), ctime(&tm));

    if (!final) printf("progress ");
    printf("one sided P value (very small numbers are bad)\n");
    printf("%c = %.3g\n", "pP"[final], pcombo(pvs, 2));
    if (!final) printf("------\n\n");
}

static short transf[256], transr[256];

static void
mktrans(void)
{
    int i, j;

    for (i=0; i<256; i++)
    {
	j=i;
	transf[i]=2*j-255;
	j=((i&1)<<7) | ((i&2)<<5) | ((i&4)<<3) | ((i&8)<<1)
		| ((i&16)>>1) | ((i&32)>>3) | ((i&64)>>5) | ((i&128)>>7);
	transr[i]=2*j-255;
    }
}

static void
dotmpacc(unsigned char *buf)
{
    int i, j;

    for (i=0; i<SIZE; i++)
    {
	j = buf[i];
	tmpacc[2*i] += transf[j];
	tmpacc[2*i+1] += transr[j];
    }
}

static void
desat(double *f, double *r)
{
    int i;

    for (i=0; i<SIZE; i++)
    {
	f[i] += tmpacc[2*i];
	tmpacc[2*i] = 0;
	r[i] += tmpacc[2*i+1];
	tmpacc[2*i+1] = 0;
    }

    tmpsat=0;
}

static void
readstuff(int64_t n, int progress)
{
    static int64_t progsize[]=
    {
	100000000,
	150000000,
	200000000,
	300000000,
	500000000,
	700000000,
	0
    };
    unsigned char buf[SIZE];
    struct poly accf, accr;
    int64_t p=0, nextp=progsize[0];
    int progi=0;

    memset(tmpacc, 0, sizeof(tmpacc));

    accf.order=SIZE; accf.coef=memmark(SIZE);
    zeropoly(accf);
    accr.order=SIZE; accr.coef=memmark(SIZE);
    zeropoly(accr);

    while (fread(buf, 1, SIZE, stdin)==SIZE)
    {
	dotmpacc(buf);
	tmpsat++;
	if (tmpsat>=120) desat(accf.coef, accr.coef);
	p++;
	if (n>=0 && p>=n) break;
	if (progress && p*SIZE>nextp)
	{
	    desat(accf.coef, accr.coef);
	    doan(p, 0, accf, accr);
	    progsize[progi] *= 10;
	    progi++;
	    nextp = progsize[progi];
	    if (!nextp) {progi = 0; nextp = progsize[progi];}
	}
    }

    desat(accf.coef, accr.coef);

    if (n>0 && p<n)
    {
fprintf(stderr, "Warning, selfcor expected %.0f 64kB blocks, saw only %.0f\n",
		(double)n, (double)p);
	seterr(1);
    }

    doan(p, 1, accf, accr);
}

int
main(int argc, char **argv)
{
    double dn;
    int64_t n= -1;
    int progress=0, j;

    tstart = time(0);

    mktrans();

    for (j=1; j<argc; j++)
    {
	if (strcmp(argv[j], "--progress")==0) progress = 1;
	else if (sscanf(argv[j], "%lf", &dn)==1)
	    {if (dn>0) n = (int64_t)(dn*(1.0/SIZE));}
	else crash("optional arg must be --progress or numeric");
    }

    readstuff(n, progress);

    return errorlevel;
}
