/* Test software for gjrand random numbers version 4.0.1.0 or later. */
/* Copyright (C) 2004-2013 G. Jones. */
/* Licensed under the GNU General Public License version 2 or 3. */

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "../../src/my_int_types.h"
#include "../../testcommon/pscale.h"

static void
crash(char *s)
{
    fprintf(stderr, "crash [%s]\n", s); exit(1);
}

static int errorlevel=0;
static void seterr(int e) {if (errorlevel<e) errorlevel=e;}

#define DIM 8
#define SIZE (3*3*3*3 * 3*3*3*3)

static const double prob[3]=
    {0.298307447927072644, 0.403385104145854712, 0.298307447927072644};

static long sat=0;
static struct nc
    {unsigned short nextbase; unsigned short scount; int32_t tots;} nc[SIZE];
static struct dc {double c, t;} count[SIZE];

static void
init(void)
{
    int j;

    for (j=0; j<SIZE; j++)
    {
	nc[j].nextbase = 3*j % SIZE; nc[j].scount = 0; nc[j].tots = 0;
	count[j].c = 0.0; count[j].t = 0.0;
    }
}

static void
desat(void)
{
    long c=0;
    int j, k;

    for (j=0; j<SIZE; j++)
    {
	k = nc[j].scount; count[j].c += k; c += k;
	count[j].t += nc[j].tots;
	nc[j].scount = 0; nc[j].tots = 0;
    }

    if (c!=sat)
    {
	fprintf(stderr, "Counters overflowed. Seriously non-random.\n");
	printf("one sided P value (very small numbers are bad)\n\nP = 0\n");
    }

    sat = 0;
}

static int
bitcount(uint32_t w)
{
    w = ((w>>1)&0x55555555) + (w&0x55555555);
    w = ((w>>2)&0x33333333) + (w&0x33333333);
    w = ((w>>4)&0x0f0f0f0f) + (w&0x0f0f0f0f);
    w = ((w>>8)&0x00ff00ff) + (w&0x00ff00ff);
    w = ((w>>16)&0x0000ffff) + (w&0x0000ffff);

    return (int)w;
}

static uint32_t
dobuf(const uint32_t *buf, int i, uint32_t st)
{
    static const int reduce[33]=
    {
	0, 0, 0, 0, 0,  0, 0, 0, 0, 0,  0, 0, 0, 0, 0,
	1, 1, 1,
	2, 2, 2, 2, 2,  2, 2, 2, 2, 2,  2, 2, 2, 2, 2
    };

    struct nc *p;
    int j, bc;

    for (j=0; j<i; j++)
    {
	p = nc+st;
	bc = bitcount(buf[j]);
	st = p->nextbase + reduce[bc];
	p->scount++;
	p->tots += ((int)bc)-16;
    }

    sat += i;
    if (sat>10000000) desat();

    return st;
}

#ifndef M_SQRT1_2
#define M_SQRT1_2 0.70710678118654752438
#endif
#define CORRECT3 0.57735026918962576451
#define CORRECT6 0.40824829046386301636

static void
mix3(double ct[], int stride)
{
    double *p1=ct+stride, *p2=p1+stride;
    double a, b, c;
    int j;

    for (j=0; j<stride; j++)
    {
	a = ct[j]; b = p1[j]; c = p2[j];
	ct[j] = (a+b+c) * CORRECT3;
	p1[j] = (2*b-a-c) * CORRECT6;
	p2[j] = (a-c) * M_SQRT1_2;
    }

    stride /= 3;
    if (stride)
    {
	mix3(ct, stride);
	mix3(p1, stride);
	mix3(p2, stride);
    }
}

static time_t tstart;

static void
doan(int64_t n, int final)
{
    double tct[SIZE];
    double pv;
    double t, x;
    time_t tm;
    int j;

    if (sat) desat();
    for (j=0; j<SIZE; j++)
    {
	if (count[j].c == 0.0) tct[j] = 0.0;
	else tct[j] = count[j].t / sqrt(count[j].c * 8.0);
    }

    if (n<6000)
    {
	fprintf(stderr, "Warning, < 24 K, don't trust result.\n");
	seterr(1);
    }

    mix3(tct, SIZE/3);

    t = -1.0;
    for (j=1; j<SIZE; j++) {x = fabs(tct[j]); if (x>t) t = x;}

    printf("mix3 extreme = %.5f\n", t);
    x = erfc(M_SQRT1_2*t);
    pv = pco_scale(x, (double)(SIZE-1));

    tm = time(0);
    printf("\nprocessed %.2g bytes in %.3g seconds. %s\n",
	n*4.0, (double)(tm-tstart), ctime(&tm));

    if (!final) printf("progress ");
    printf("one sided P value (very small numbers are bad)\n");
    printf("%c = %.3g\n", "pP"[final], pv);
    if (!final) printf("------\n\n");
}

static unsigned
mktrans(unsigned char buf[], unsigned ts)
{
    unsigned k;
    int j;

    for (j=0; j<4096; j++)
    {
	k = buf[j];
	buf[j] = (k>>1) ^ k ^ ts;
	ts = (k&1)<<7;
    }

    return ts;
}

static void
readstuff(int64_t n, int trans, int progress)
{
    static int64_t progsize[]=
    {
	100000000,
	150000000,
	200000000,
	300000000,
	500000000,
	700000000,
	0
    };
    uint32_t buf[1024];
    int64_t pos=0, nextp=progsize[0];
    uint32_t st=538;
    int j, progi=0;
    unsigned ts=0;

    while (n<0 || pos<n)
    {
	j = fread(buf, 4, 1024, stdin);
	if (j<=0) break;
	if (trans) ts = mktrans((unsigned char *)buf, ts);
	if (n>0 && j>n-pos) j = n-pos;
	st = dobuf(buf, j, st);
	pos += j;
	if (j<1024) break;
	if (progress && pos*4>nextp)
	{
	    doan(pos, 0);
	    progsize[progi] *= 10;
	    progi++;
	    nextp = progsize[progi];
	    if (!nextp) {progi = 0; nextp = progsize[progi];}
	}
    }

    if (n>0 && pos<n)
    {
	fprintf(stderr, "Warning, expected %.0f bytes, saw only %.0f\n",
		(double)n, (double)pos);
	seterr(1);
    }

    doan(pos, 1);
}

int
main(int argc, char **argv)
{
    double dn;
    int64_t n= -1;
    int trans=0, progress=0, j;

    tstart = time(0);

    init();

    for (j=1; j<argc; j++)
    {
	if (strcmp(argv[j], "--progress")==0) progress = 1;
	else if (strcmp(argv[j], "-t")==0) trans = 1;
	else if (sscanf(argv[j], "%lf", &dn)==1) n = (int64_t)(dn*0.25);
	else crash("optional arg must be --progress or -t or numeric");
    }

    readstuff(n, trans, progress);

    return errorlevel;
}
