/* Test software for gjrand random numbers version 3.4.1.0 or later. */
/* Copyright (C) 2004-2011 G. Jones. */
/* Licensed under the GNU General Public License version 2 or 3. */

/* This is tests related to the balance of 0 and 1 bits in the test data. */
/* New in version 1.0.0.0 . Fifth moments in 1.3.0.0 . */

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "../../src/my_int_types.h"
#include "../../testcommon/pcombo.h"

static void
crash(char *s)
{
    fprintf(stderr, "crash [%s]\n", s); exit(1);
}

static int errorlevel=0;
static void seterr(int e) {if (errorlevel<e) errorlevel=e;}

static double totone=0.0;
static double var[9]={0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
static double skew[9]={0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
static double five[9]={0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};

static double xtot, xvar[9], xskew[9], xfive[9];

static double *
mkex1(int coins, int64_t count)
{
    static double p1[32771], p2[32771];
    double *x1=p1, *x2=p2, *tmp, t;
    int i, j;
    int hi=0, lo=0;

    for (i=0; i<=coins; i++) x1[i]=x2[i]=0.0;
    x1[0]=count; /* with zero coins thrown, zero heads always results */

    for (i=1; i<=coins; i++)
    {
	for (j=lo; j<=hi; j++) {t=x1[j]*0.5; x2[j]+=t; x2[j+1]=t;}
	hi++;
	for (j=lo; j<=hi; j++) x1[j]=0.0;
	while (lo<hi && x2[lo]<1e-30) lo++;
	while (hi>lo && x2[hi]<1e-30) hi--;

	tmp=x1; x1=x2; x2=tmp;
    }

    return x1;
}

static double
getvar(double *p, int coins, int64_t count)
{
    double v=coins*0.25;
    double r=0.0;
    int i, mx=coins/2;

    for (i=0; i<=coins; i++)
    {
	double t=i-mx;
	t*=t;
	t-=v;
	t*=t;
	r+=p[i]*t;
    }

    return sqrt(r)/count;
}

static double
getskew(double *p, int coins, int64_t count)
{
    double r=0.0;
    int i, mx=coins/2;

    for (i=0; i<=coins; i++)
    {
	double t=i-mx;
	t*=t*t;
	t*=t;
	r+=p[i]*t;
    }

    return sqrt(r)/count;
}

static double
getfive(double *p, int coins, int64_t count)
{
    double r=0.0;
    int i, mx=coins/2;

    for (i=0; i<=coins; i++)
    {
	double t=i-mx;
	t*=(t*t)*(t*t);
	t*=t;
	r+=p[i]*t;
    }

    return sqrt(r)/count;
}

static void
mkexpect(int64_t pos)
{
    double *p;
    int i;

    xtot=sqrt(pos*8.0)*0.5;
    for (i=0; i<9; i++)
    {
	p=mkex1(128<<i, pos/(16<<i));
	xvar[i]=getvar(p, 128<<i, pos/(16<<i));
	xskew[i]=getskew(p, 128<<i, pos/(16<<i));
	xfive[i]=getfive(p, 128<<i, pos/(16<<i));
    }
}

static time_t tstart;

static void
doan(int64_t n, int final)
{
    double pm, pvar[9], pskew[9], pfive[9], x, y, z;
    time_t tm;
    int i;

    mkexpect(n);

    x=totone/xtot;
    pm=erfc(fabs(M_SQRT1_2*x));
    printf("total bitcount discrepency %.3f sigma, (p = %.3g)\n", x, pm);

    printf("block   var/sigma	  p    skew/sigma          p    five/sigma          p\n");
    for (i=0; i<9; i++)
    {
	x=var[i]*(16<<i)/n;
	x-=(32<<i);
	x/=xvar[i];
	pvar[i]=erfc(fabs(M_SQRT1_2*x));

	y=skew[i]*(16<<i)/n;
	y/=xskew[i];
	pskew[i]=erfc(fabs(M_SQRT1_2*y));

	z=five[i]*(16<<i)/n;
	z/=xfive[i];
	pfive[i]=erfc(fabs(M_SQRT1_2*z));

	printf("%5d: %10.3f (%9.3g) ; %10.3f (%9.3g) ; %10.3f (%9.3g)\n",
		128<<i, x, pvar[i], y, pskew[i], z, pfive[i]);
    }
    pvar[0]=pcombo(pvar, 9);
    pvar[1]=pcombo(pskew, 9);
    pvar[2]=pcombo(pfive, 9);
    pvar[3]=pm;
    printf("(%.3g, %.3g, %.3g)\n", pvar[0], pvar[1], pvar[2]);

    tm = time(0);
    printf("\nprocessed %.2g bytes in %.3g seconds. %s\n",
	(double)n, (double)(tm-tstart), ctime(&tm));

    if (!final) printf("progress ");
    printf("one sided P value (very small numbers are bad)\n");
    printf("%c = %.3g\n", "pP"[final], pcombo(pvar, 4));
    if (!final) printf("------\n\n");
}

static double
do16(unsigned char * b)
{
    double r, s, c;
    uint32_t * const x=(uint32_t * const)b;
    uint32_t v=x[0], w=x[1], t=x[2], u=x[3];

    v = ((v>>1)&0x55555555) + (v&0x55555555);
    w = ((w>>1)&0x55555555) + (w&0x55555555);
    t = ((t>>1)&0x55555555) + (t&0x55555555);
    u = ((u>>1)&0x55555555) + (u&0x55555555);

    v = ((v>>2)&0x33333333) + (v&0x33333333);
    w = ((w>>2)&0x33333333) + (w&0x33333333);
    t = ((t>>2)&0x33333333) + (t&0x33333333);
    u = ((u>>2)&0x33333333) + (u&0x33333333);
    v += t; w += u;

    t = v>>4; u = w>>4; v &= 0x0f0f0f0f; w &= 0x0f0f0f0f;
    t &= 0x0f0f0f0f; u &= 0x0f0f0f0f; v+=t; w+=u;
    v+=w;

    w = v>>8; v &= 0x00ff00ff; w &= 0x00ff00ff; v+=w;

    w = v>>16; v &= 0x0000ffff; v+=w;

    r=((int)v)-64;

    s=r*r;
    var[0]+=s;
    c=s*r;
    skew[0]+=c;
    five[0]+=c*s;

    return r;
}

static double
don(unsigned char *b, int sz, int lv)
{
    double r, s, c;

    lv--;
    if (lv<=0) {r=do16(b); r+=do16(b+16);}
    else
    {
	sz >>= 1;
	r=don(b, sz, lv); r+=don(b+sz, sz, lv);
    }
    lv++;

    s=r*r;
    var[lv]+=s;
    c=s*r;
    skew[lv]+=c;
    five[lv]+=c*s;
    return r;
}

static void
do4096(unsigned char *b)
{
    totone += don(b, 4096, 8);
}

static unsigned
mktrans(unsigned char buf[], unsigned ts)
{
    unsigned k;
    int j;

    for (j=0; j<4096; j++)
    {
	k = buf[j];
	buf[j] = (k>>1) ^ k ^ ts;
	ts = (k&1)<<7;
    }

    return ts;
}

static void
readstuff(int64_t n, int trans, int progress)
{
    static int64_t progsize[]=
    {
	100000000,
	150000000,
	200000000,
	300000000,
	500000000,
	700000000,
	0
    };
    uint32_t buf[1024];
    int64_t pos=0, nextp=progsize[0];
    int j, progi=0;
    unsigned ts=0;

    while (n<0 || pos<n)
    {
	j=fread(buf, 4096, 1, stdin);
	if (j!=1) break;
	if (trans) ts = mktrans((unsigned char *)buf, ts);
	pos += 4096;
	do4096((unsigned char *)buf);
	if (progress && pos>nextp)
	{
	    doan(pos, 0);
	    progsize[progi] *= 10;
	    progi++;
	    nextp = progsize[progi];
	    if (!nextp) {progi = 0; nextp = progsize[progi];}
	}
    }

    if (n>0 && pos<n)
    {
	fprintf(stderr, "Warning, expected %.0f bytes, saw only %.0f\n",
		(double)n, (double)pos);
	seterr(1);
    }

    if (pos<50000) /* HERE This is still a guess. But mostly tests ok. */
    {
	fprintf(stderr, "Warning, < 50000 bytes, don't trust result.\n");
	seterr(1);
    }

    if (trans) printf("transitions\n");

    doan(pos, 1);
}

int
main(int argc, char **argv)
{
    double dn;
    int64_t n= -1;
    int trans=0, progress=0, j;

    tstart = time(0);

    for (j=1; j<argc; j++)
    {
	if (strcmp(argv[j], "--progress")==0) progress = 1;
	else if (strcmp(argv[j], "-t")==0) trans = 1;
	else if (sscanf(argv[j], "%lf", &dn)==1) n = (int64_t)dn;
	else crash("optional arg must be --progress or -t or numeric");
    }

    if (n>0) n &= ~((int64_t)4095);

    readstuff(n, trans, progress);

    return errorlevel;
}
