/* Test software for gjrand random numbers version 3.4.1.0 or later. */
/* Copyright (C) 2004-2011 G. Jones. */
/* Licensed under the GNU General Public License version 2 or 3. */

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "../../src/my_int_types.h"
#include "../../testcommon/pcombo.h"
#include "../../testcommon/binom.h"
#include "../../testcommon/chi2p.h"

static void
crash(char *s)
{
    fprintf(stderr, "crash [%s]\n", s); exit(1);
}

static int errorlevel=0;
static void seterr(int e) {if (errorlevel<e) errorlevel=e;}

static int64_t count[6][256];
static int32_t last[6];
static double prob[6];
static const char counts[6] = {9, 13, 15, 17, 19, 23};
static const char types[6] = {0, 1, 1, 1, 1, 0};
static const int dfs[2] = {200, 16};

static void
newline(const double from[], double to[], int fromcount)
{
    int j;

    to[0] = from[0]*0.5;
    for (j=1; j<=fromcount; j++) to[j] = (from[j-1]+from[j])*0.5;
    to[fromcount+1] = from[fromcount]*0.5;
}

static void
pasc(double *pr, int order)
{
    double t[100];
    int j;

    pr[0] = 1.0;

    for (j=0; j<order; j++)
    {
	newline(pr, t, j);
	j++; if (j>=order) break;
	newline(t, pr, j);
    }
}

static void
init(void)
{
    double pr[65];
    double cu[32];
    double acc;
    int j;

    memset(count, 0, sizeof(count));

    pasc(pr, 32);
    acc = 0.0;
    for (j=0; j<16; j++) {acc += pr[j]; cu[j] = acc;}
    for (j=0; j<3; j++) prob[5-j] = prob[j] = cu[(int)(counts[j])];

    /* This is very approximate. But for decently large files */
    /* it has very little effect anyway. */
    for (j=0; j<6; j++) last[j] = 1024-((int)(1.0/prob[j]))-1;
}

static void
dogap(int b, int32_t pos)
{
    int g;

    if (b<=15)
    {
	g = pos-last[2];
	if (g>255) g = 0;
	count[2][g]++;
	last[2] = pos;
	if (b<=13)
	{
	    g = pos-last[1];
	    if (g>255) g = 0;
	    count[1][g]++;
	    last[1] = pos;
	    if (b<=9)
	    {
		g = pos-last[0];
		if (g>255) g = 0;
		count[0][g]++;
		last[0] = pos;
	    }
	}
    }
    else if (b>=17)
    {
	g = pos-last[3];
	if (g>255) g = 0;
	count[3][g]++;
	last[3] = pos;
	if (b>=19)
	{
	    g = pos-last[4];
	    if (g>255) g = 0;
	    count[4][g]++;
	    last[4] = pos;
	    if (b>=23)
	    {
		g = pos-last[5];
		if (g>255) g = 0;
		count[5][g]++;
		last[5] = pos;
	    }
	}
    }
}

static void
doword(uint32_t w, int32_t pos)
{
    w = ((w>>1)&0x55555555) + (w&0x55555555);
    w = ((w>>2)&0x33333333) + (w&0x33333333);
    w = ((w>>4)&0x0f0f0f0f) + (w&0x0f0f0f0f);
    w = ((w>>8)&0x00ff00ff) + (w&0x00ff00ff);
    w = ((w>>16)&0x0000ffff) + (w&0x0000ffff);

    dogap(w, pos);
}

static void
dobuf(const uint32_t *buf, int i)
{
    int j;

    for (j=0; j<i; j++) doword(buf[j], j);
}

static int
mkexpect(double p, int df, double expect[256])
{
    double x=1.0, t;
    int j;

    for (j=1; j<=df; j++)
    {
	t = x*p;
	expect[j] = t;
	x -= t;
    }

    expect[0] = x;

    return j;
}

static void
doan1(int offset, double pv[2])
{
    double expect[256];
    int64_t ct[256];
    double chis, x, t, pval, extr = 999.9;
    int64_t l, tot;
    int type = types[offset];
    int df = dfs[type];
    int j, exi = -1, exd = '?';

    mkexpect(prob[offset], df, expect);

    memcpy(ct, &(count[offset][0]), sizeof(ct));

    for (j=df+1; j<256; j++) ct[0] += ct[j];
    l = 0;
    for (j=0; j<=df; j++) l += ct[j];
    if (l!=0) {t = l; for (j=0; j<=df; j++) expect[j] *= t;}

    chis = 0.0;

    for (j=0; j<=df; j++)
    {
	x = expect[j];
	t = ct[j]-x;
	chis += t*t/x;
    }

    pval = chi2p1(chis, df);
    pv[0] = pval;

    tot = l;
    for (j=1; j<=df; j++)
    {
	int64_t k = ct[j];
	x = sumbino(tot, k, prob[offset]);
	if (x<extr)
	{
	    extr = x; exi = j;
	    if (k>tot*prob[offset]) exd = '+'; else exd = '-';
	}
	tot -= k;
    }

    extr = pco_scale(extr, df);

    pv[1] = extr;

    printf("%2d %8.5f %3d %12.3f %12.6g : (%3d %c) %12.6g\n",
	counts[offset], prob[offset], df, chis, pval, exi, exd, extr);
}

static time_t tstart;

static void
doan(int64_t n, int final)
{
    double pv[12];
    time_t tm;
    int j;

    for (j=0; j<6; j++) doan1(j, pv+2*j);

    tm = time(0);
    printf("\nprocessed %.2g bytes in %.3g seconds. %s\n",
	n*4.0, (double)(tm-tstart), ctime(&tm));

    if (!final) printf("progress ");
    printf("one sided P value (very small numbers are bad)\n");
    printf("%c = %.3g\n", "pP"[final], pcombo(pv, 12));
    if (!final) printf("------\n\n");
}

static void
agelast(void)
{
    int j, k;

    for (j=0; j<6; j++)
    {
	k = last[j]-1024;
	if (k < -20000) k = -20000;
	last[j] = k;
    }
}

static unsigned
mktrans(unsigned char buf[], unsigned ts)
{
    unsigned k;
    int j;

    for (j=0; j<4096; j++)
    {
	k = buf[j];
	buf[j] = (k>>1) ^ k ^ ts;
	ts = (k&1)<<7;
    }

    return ts;
}

static void
readstuff(int64_t n, int trans, int progress)
{
    static int64_t progsize[]=
    {
	100000000,
	150000000,
	200000000,
	300000000,
	500000000,
	700000000,
	0
    };
    uint32_t buf[1024];
    int64_t pos=0, nextp=progsize[0];
    int j, progi=0;
    unsigned ts=0;

    while (n<0 || pos<n)
    {
	j=fread(buf, 4, 1024, stdin);
	if (j<=0) break;
	if (trans) ts = mktrans((unsigned char *)buf, ts);
	if (n>0 && j>n-pos) j=n-pos;
	agelast();
	dobuf(buf, j);
	pos+=j;
	if (j<1024) break;
	if (progress && pos*4>nextp)
	{
	    doan(pos, 0);
	    progsize[progi] *= 10;
	    progi++;
	    nextp = progsize[progi];
	    if (!nextp) {progi = 0; nextp = progsize[progi];}
	}
    }

    if (trans) printf("transitions\n");

    if (n>0 && pos<n)
    {
	fprintf(stderr, "Warning, expected %.0f bytes, saw only %.0f\n",
		(double)n, (double)pos);
	seterr(1);
    }

    if (4*pos<1900000)
    {
	fprintf(stderr, "Warning, < 1900000 bytes, don't trust result.\n");
	seterr(1);
    }

    doan(pos, 1);
}

int
main(int argc, char **argv)
{
    double dn;
    int64_t n= -1;
    int trans=0, progress=0, j;

    tstart = time(0);

    init();

    for (j=1; j<argc; j++)
    {
	if (strcmp(argv[j], "--progress")==0) progress = 1;
	else if (strcmp(argv[j], "-t")==0) trans = 1;
	else if (sscanf(argv[j], "%lf", &dn)==1) n = (int64_t)(dn*0.25);
	else crash("optional arg must be --progress or -t or numeric");
    }

    readstuff(n, trans, progress);

    return errorlevel;
}
